"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

from aurori import app_router
import time

# index route
#@app_router.get('/')
#def index():
#    return {}

#@app_router.get('/<path:path>')
#def static_proxy(path):
#    return {}

'''
@app_bp.after_request
def add_header(r):

    r.headers[
        "Cache-Control"] = "no-cache, no-store, must-revalidate, max-age=0"
    r.headers["Pragma"] = "no-cache"
    r.headers["Expires"] = "0"
    return r
'''



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

from datetime import datetime
from aurori.logs import log_manager
from aurori.config import config_manager
from aurori.database import db
import logging
import traceback


class Job(object):
    """Base class that each job inherit from.
       The class define methods that all jobs have to implement
    """
    description = "Not available"  # description of the job
    disable = False  # disable the job
    local = True  # only runnable by the same feature
    strict = True  # strict argument parsing
    requireAdmin = False  # admin is required to view the page
    requirePermission = None  # a permission is required in the meaning of one of the following

    # Repetetive members
    cron = False
    day = None
    week = None
    day_of_week = None
    hour = None
    minute = None
    second = None

    def __init__(self, name=None, uri=None):
        if name is None:
            self.name = type(self).__name__
        else:
            self.name = name
        self.job_key = self.name
        self.feature = ""
        self.parameters = None
        self.defineArguments()

    def start_job(self, **kwargs):
        from aurori.jobs.database import JobExecute

        self.debug_info = {}
        triggered = datetime.now()
        current_time = triggered.strftime("%H:%M:%S")

        log_manager.info("Run " + self.name + " " + current_time)

        je = None

        with db.get_session() as db_session:
            if 'job_execution_id' in kwargs:
                je = db_session.query(JobExecute).filter_by(
                    id=kwargs['job_execution_id']).first()

            if je is None:
                je = JobExecute()
                je.feature = self.feature
                je.triggered_by = "Cron"
                je.triggered_on = triggered
                je.name = self.job_key
                db_session.add(je)
                db_session.commit()

            try:
                self.run(**kwargs)
                je.state = "SUCCEED"
                # je.results = {"hjsadhj": "jklsajdklas"}
            except Exception as e:
                je.state = "FAILED"
                je.results = {
                    "error": str(type(e)),
                    "traceback": str(traceback.format_exc()),
                    "debug": self.debug_info
                }
                print(je.results)

            after = datetime.now()
            delta = after - triggered
            je.lifetime = delta.total_seconds()
            db_session.commit()


    def addArgument(self,
                    name,
                    typestring,
                    label="",
                    description="",
                    optional=False,
                    group=None):
        p = {
            'name': name,
            'type': typestring,
            'label': label,
            'optional': optional,
            'group': group
        }
        if self.parameters is None:
            self.parameters = [p]
        else:
            self.parameters.append(p)

    def addDictArgument(self,
                        name,
                        label="",
                        description="",
                        optional=False,
                        group=None):
        log_manager.info("Add dict type argument to job {}".format(self.name))
        self.addArgument(name, 'dict', label, description, optional, group)

    def addListArgument(self,
                        name,
                        label="",
                        description="",
                        optional=False,
                        group=None):
        log_manager.info("Add list type argument to job {}".format(self.name))
        self.addArgument(name, 'list', label, description, optional, group)

    def addStringArgument(self,
                          name,
                          label="",
                          description="",
                          optional=False,
                          group=None):
        log_manager.info("Add string type argument to job {}: {} - {}".format(
            self.name, name, label))
        self.addArgument(name, 'string', label, description, optional, group)

    def addDoubleArgument(self,
                          name,
                          label="",
                          description="",
                          optional=False,
                          group=None):
        log_manager.info("Add double type argument to job {}".format(self.name))
        self.addArgument(name, 'double', label, description, optional, group)

    def addIntegerArgument(self,
                           name,
                           label="",
                           description="",
                           optional=False,
                           group=None):
        log_manager.info("Add integer type argument for job {}".format(
            self.name))
        self.addArgument(name, 'integer', label, description, optional, group)

    def addDatetimeArgument(self,
                            name,
                            label="",
                            description="",
                            optional=False,
                            group=None):
        log_manager.info("Add datetime type argument for job {}".format(
            self.name))
        self.addArgument(name, 'datetime', label, description, optional, group)

    def addTimeArgument(self,
                        name,
                        label="",
                        description="",
                        optional=False,
                        group=None):
        log_manager.info("Add time type argument for job {}".format(self.name))
        self.addArgument(name, 'time', label, description, optional, group)

    def addDateArgument(self,
                        name,
                        label="",
                        description="",
                        optional=False,
                        group=None):
        log_manager.info("Add date tyme argument for job {}".format(self.name))
        self.addArgument(name, 'date', label, description, optional, group)

    def addBooleanArgument(self,
                           name,
                           label="",
                           description="",
                           optional=False,
                           group=None):
        log_manager.info("Add boolean type argument for job {}".format(
            self.name))
        self.addArgument(name, 'boolean', label, description, optional, group)

    def defineArguments(self):
        pass

    def run(self, *args, **kwargs):
        raise NotImplementedError



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import hashlib
import bcrypt
from datetime import datetime
from sqlalchemy import Boolean, Column, DateTime, Integer, String, LargeBinary
from sqlalchemy.ext.hybrid import hybrid_property, hybrid_method

from aurori.database import SQLModelBase
from aurori.utils import unicode_string_to_bytes


class UserBase(SQLModelBase):
    __abstract__ = True
    __tablename__ = 'users'
    # nonvolatile data stored in the db
    id = Column(Integer, primary_key=True, autoincrement=True)
    _password_hash = Column(LargeBinary(60), nullable=False)
    _pin_hash = Column(LargeBinary(60))
    _salt = Column(String(128))
    _api_key_hash = Column(String(256), nullable=True, default=None)
    email = Column(String(120), index=True, unique=True)
    username = Column(String(120), index=True, unique=True)
    firstname = Column(String(120), default="")
    lastname = Column(String(120), default="")
    organization = Column(String(120), default="")
    phone = Column(String(64), default="")
    account_created_date = Column(DateTime, default=datetime.utcnow)
    last_login_date = Column(DateTime, default=datetime.utcnow)
    password_reset_expired_date = Column(DateTime, default=datetime.utcnow)
    password_reset_hash = Column(String(128), default="")
    account_verified = Column(Boolean, default=False)
    account_activated = Column(Boolean, default=False)
    account_locked = Column(Boolean, default=False)
    ldap = Column(Boolean, default=False)
    ldap_dn = Column(String(64), default="")
    ldap_username = Column(String(64), default="")
    admin = Column(Boolean, default=False)
    pinAttempts = Integer()
    loginAttempts = Integer()

    def __init__(self, email, password, isAdmin=False, skip_password=False):
        self.email = email
        if skip_password is False:
            self.password = password
        self.admin = isAdmin

    def __repr__(self):
        return '<User {} {} {}>'.format(self.firstname, self.lastname,
                                          self.email)

    @hybrid_property
    def password(self):
        return self._password_hash

    @password.setter
    def password(self, plaintext_password):
        password_bytes = unicode_string_to_bytes(plaintext_password)
        salt = bcrypt.gensalt(rounds=12, prefix=b'2b')
        self._password_hash =  bcrypt.hashpw(password_bytes, salt)

    @hybrid_method
    def checkPassword(self, plaintext_password):
        return bcrypt.checkpw(unicode_string_to_bytes(plaintext_password),self.password)

    @hybrid_property
    def pin(self):
        return self._pin_hash

    @pin.setter
    def pin(self, plaintext_pin):
        pin_bytes = unicode_string_to_bytes(plaintext_pin)
        salt = bcrypt.gensalt(rounds=12, prefix=b'2b')
        self._password_hash =  bcrypt.hashpw(pin_bytes, salt)

    @hybrid_method
    def checkPin(self, plaintext_pin):
        return bcrypt.checkpw(unicode_string_to_bytes(plaintext_pin),self.pin)


    @hybrid_property
    def api_key(self):
        return self._api_key_hash

    @api_key.setter
    def api_key(self, plaintext_api_key):
        self._api_key_hash = hashlib.sha512(
            plaintext_api_key.encode('utf8')).hexdigest()

    @hybrid_method
    def check_api_key(self, plaintext_api_key):
        return hashlib.sha512(plaintext_api_key).hexdigest() == self.api_key

    # checks if user has a certain permission
    @hybrid_method
    def has_permission(self, permission_to_check):
        for permission_group in self.permission_groups:
            for permission in permission_group.permissions:
                if type(permission_to_check) is str:
                    permission_key = permission_to_check
                else:
                    permission_key = str(permission_to_check.__module__).split(
                        '.')[1] + '.' + permission_to_check.__name__
                if permission.name == permission_key:
                    return True
        return False



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import inspect
import pkgutil
import sqlalchemy as sa

from aurori.features.page import Page
from aurori.features.permission import Permission
from aurori.jobs.job import Job
from aurori.jobs import job_manager
from aurori.logs import log_manager
from fastapi import APIRouter


class Feature(object):
    """Base class that each feature inherit from.
       The class define methods that all features have to implement
    """

    disabled = False

    def __init__(self, app, db, name=None, uri=None):
        self.description = 'UNKNOWN'
        if name is None:
            self.name = type(self).__name__
        else:
            self.name = name
        if uri is None:
            self.uri = self.name
        else:
            self.uri = uri
        self.permissions = {}
        self.dataViews = {}
        self.sections = {}
        self.pages = {}
        self.actions = []
        self.app = app
        self.db = db

    def addSinglePageSection(self, name, page, caption=None, permission=None):
        key = self.name + "_" + name
        if caption is None:
            caption = name
        self.sections[key] = {
            "caption": caption,
            "page": page,
            "permission": permission
        }

    def addPage(self, name, callback, caption=None, permission=None):
        key = self.name + "_" + name
        if caption is None:
            caption = name
        self.pages[key] = {
            "caption": caption,
            "callback": callback,
            "permission": permission
        }

    def get_page(self, name):
        key = self.name + "_" + name
        return self.pages[key]

    def discover_jobs(self, feature_source_path):
        jobSource = feature_source_path + '.' + self.name + '.jobs'
        imported_source = __import__(jobSource, fromlist=['blah'])

        for _, jobname, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                job_module = __import__(jobname, fromlist=['blah'])
                clsmembers = inspect.getmembers(job_module, inspect.isclass)
                for (_, c) in clsmembers:
                    if issubclass(c, Job) & (c is not Job):
                        if c.disable is False:
                            log_manager.info(
                                f'    Job registered from "{c.__module__}"')
                            job_manager.register_job(self, c, True)

    def discover_apis(self, feature_source_path):
        feature_api_path = feature_source_path + '.' + self.name + '.api'
        api_source = __import__(feature_api_path, fromlist=['blah'])
        for _, modulename, ispkg in pkgutil.iter_modules(
                api_source.__path__, api_source.__name__ + '.'):
            if not ispkg:
                api_module = __import__(modulename,
                                               fromlist=['blah'])
                clsmembers = inspect.getmembers(api_module)
                for (api_name, c) in clsmembers:
                    if isinstance(c, APIRouter):
                        log_manager.info(
                            f'    API "{api_name}" loaded from "{modulename}"'
                        )
                        if c.prefix != "":
                            self.app.include_router(c, tags=["App." + self.name])
                        else:
                            self.app.include_router(c,prefix= "/" + self.uri, tags=["App." + self.name])


    def discover_permissions(self, feature_source):
        permissionsSource = feature_source + '.' + self.name
        imported_source = __import__(permissionsSource, fromlist=['blah'])

        for _, permissionname, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                permission_module = __import__(permissionname,
                                               fromlist=['blah'])
                clsmembers = inspect.getmembers(permission_module,
                                                inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for Permission classes
                    if issubclass(c, Permission) & (c is not Permission):
                        permissionInstance = c()
                        log_manager.info(
                            f'    Permission "{permissionInstance.name}" loaded from "{c.__module__}"'
                        )
                        self.add_permission(str(permissionInstance.name),
                                           permissionInstance)

    def discoverSections(self, feature_source_path):
        '''
        sectionsSource = feature_source_path + '.' + self.name
        imported_source = __import__(sectionsSource, fromlist=['blah'])

        for _, sectionname, ispkg in pkgutil.iter_modules(imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                section_module = __import__(sectionname, fromlist=['blah'])
                clsmembers = inspect.getmembers(section_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for Section classes
                    if issubclass(c, Section) & (c is not Section):
                        sectionInstance = c()
                        log_manager.info(f'Section: "{c.__module__}" created for feature: "{self.name}"')
                        #self.dataViews[str(dataViewInstance.uri)] = (dataViewInstance)
        '''

    def discoverNodeClasses(self, feature_source_path):
        # for _, node_class, ispkg in pkgutil.iter_modules(imported_source.__path__, imported_source.__name__ + '.'):
        #    if ispkg:
        #        print(node_class)

        node_sources = feature_source_path + '.' + self.name + '.' + 'nodes'
        imported_source = __import__(node_sources, fromlist=['blah'])

        for _, nodename, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                node_module = __import__(nodename, fromlist=['blah'])
                clsmembers = inspect.getmembers(node_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for DataView classes
                    if issubclass(c, NodeClass) & (c is not NodeClass):
                        if c.disable is False:
                            log_manager.info(
                                f'NodeClass: "{c.__module__}" loaded from "{node_sources}"'
                            )
                            nodeManager.register_node_class(self, c)
                        else:
                            log_manager.info(
                                f'NodeClass: "{c.__module__}" is disabled and wont show up"'
                            )
        '''
        for i in os.listdir(node_sources):
            module_path = os.path.join(node_sources,i)
            if os.path.isdir(module_path):
                module_path_list = module_path.split('/')
                module_path_list.append('node')
                node_module_source = '.'.join(module_path_list)
                node_module = __import__(node_module_source, fromlist=['blah'])
                clsmembers = inspect.getmembers(node_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for DataView classes
                    if issubclass(c, NodeClass) & (c is not NodeClass):
                        if c.disable is False:
                            log_manager.info(f'NodeClass: "{c.__module__}" loaded from "{node_sources}"')
                            nodeManager.register_node_class(self, c)
                        else:
                            log_manager.info(f'NodeClass: "{c.__module__}" is disabled and wont show up"')
        '''

    def discover_pages(self, feature_source_path):
        pagesSource = feature_source_path + '.' + self.name + '.' + 'pages'
        imported_source = __import__(pagesSource, fromlist=['blah'])

        for _, pagename, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                page_module = __import__(pagename, fromlist=['blah'])
                clsmembers = inspect.getmembers(page_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for DataView classes
                    if issubclass(c, Page) & (c is not Page):
                        pageInstance = c()
                        log_manager.info(
                            f'    Page: "{pageInstance.name}" loaded from "{c.__module__}"'
                        )
                        if pageInstance.disable is True:
                            log_manager.info(
                                f'    Page: "{pageInstance.name}" is disabled and wont show up'
                            )
                        else:
                            self.pages[str(pageInstance.name)] = pageInstance

    def discoverDataViews(self, feature_source_path):
        actionsSource = feature_source_path + '.' + self.name + '.' + 'views'
        imported_source = __import__(actionsSource, fromlist=['blah'])

        for _, actionname, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                action_module = __import__(actionname, fromlist=['blah'])
                clsmembers = inspect.getmembers(action_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for DataView classes
                    if issubclass(c, DataView) & (c is not DataView):
                        if hasattr(c, 'name'):
                            dataViewInstance = c(c.name, c.uri)
                            log_manager.info(
                                f'DataView "{c.name}" loaded from "{c.__module__}"'
                            )
                        else:
                            dataViewInstance = c(c.__name__, c.uri)
                            log_manager.info(
                                f'DataView "{c.__name__}" loaded from "{c.__module__}"'
                            )

                        dataViewInstance.defineProperties()
                        dataViewInstance.defineMetadata()

                        if dataViewInstance.disable is True:
                            log_manager.info(
                                f'DataView: "{dataViewInstance.name}" is disabled and wont show up'
                            )
                        else:
                            self.dataViews[str(
                                dataViewInstance.uri)] = (dataViewInstance)

                        if dataViewInstance.entrykey is None:
                            raise LookupError(
                                "DataView {} dont define a key".format(
                                    dataViewInstance.name))

    def discoverActions(self, feature_source_path):

        # create manager wide generic handlers for the view
        self.getViewActionHandler = viewHandler.GetViewActionHandler(
            self.app, self.db)
        self.createViewEntryActionHandler = viewHandler.CreateViewEntryActionHandler(
            self.app, self.db)
        self.removeViewEntryActionHandler = viewHandler.RemoveViewEntryActionHandler(
            self.app, self.db)
        self.updateViewEntryActionHandler = viewHandler.UpdateViewEntryActionHandler(
            self.app, self.db)
        self.executeViewActionsActionHandler = viewHandler.ExecuteViewActionsActionHandler(
            self.app, self.db)

        # add handlers for view actions
        # the view handlers are generic and added to every feature
        self.actions.append(self.getViewActionHandler)
        self.actions.append(self.createViewEntryActionHandler)
        self.actions.append(self.removeViewEntryActionHandler)
        self.actions.append(self.updateViewEntryActionHandler)
        self.actions.append(self.executeViewActionsActionHandler)

        log_manager.info(
            f'Actions for handling DataViews created for feature "{self.name}"'
        )

        actionsSource = feature_source_path + '.' + self.name + '.' + 'actions'
        # print("Discover actions for", self.uri, "from", actionsSource)
        imported_source = __import__(actionsSource, fromlist=['blah'])

        # print(imported_source)

        for _, actionname, ispkg in pkgutil.iter_modules(
                imported_source.__path__, imported_source.__name__ + '.'):
            if not ispkg:
                action_module = __import__(actionname, fromlist=['blah'])
                clsmembers = inspect.getmembers(action_module, inspect.isclass)
                for (_, c) in clsmembers:
                    # Check for Action classes
                    if issubclass(c, Action) & (c is not Action):
                        actionInstance = c(self.app)
                        log_manager.info(
                            f'Action "{c.__module__}" created for feature "{self.name}"'
                        )
                        if actionInstance.disable is True:
                            log_manager.info(
                                f'Action: "{actionInstance.name}" is disabled and wont show up'
                            )
                        else:
                            self.actions.append(actionInstance)

    def discoverCommands(self, feature_source_path):
        commandSource = feature_source_path + '.' + self.name + '.' + 'commands'
        try:
            imported_source = __import__(commandSource, fromlist=['blah'])
            for _, commandname, ispkg in pkgutil.iter_modules(
                    imported_source.__path__, imported_source.__name__ + '.'):
                if not ispkg:
                    try:
                        command_module = __import__(commandname,
                                                    fromlist=['blah'])
                        command_module.register(imported_source.user_cli)
                        log_manager.info(
                            'Custom command "{} > {}" registered.'.format(
                                imported_source.user_cli.name,
                                list(imported_source.user_cli.commands)[-1],
                            ))
                    except Exception as e:
                        log_manager.error(
                            'Unable to import: {}, exception: {}'.format(
                                commandname, e))

            self.app.cli.add_command(imported_source.user_cli)
        except Exception:
            pass

    def discoverWidgets(self, feature_source_path):
        widgetSource = feature_source_path + '.' + self.name + '.' + 'widgets'
        try:
            imported_source = __import__(widgetSource, fromlist=['blah'])
            for _, widgetname, ispkg in pkgutil.iter_modules(
                    imported_source.__path__, imported_source.__name__ + '.'):
                if not ispkg:
                    widget_module = __import__(widgetname, fromlist=['blah'])
                    clsmembers = inspect.getmembers(widget_module,
                                                    inspect.isclass)
                    for (_, c) in clsmembers:
                        if issubclass(c, Widget) & (c is not Widget):
                            if c.disable is False:
                                log_manager.info(
                                    f'Widget registered from "{c.__module__}"')
                                widgetInstance = widgetManager.register_widget(
                                    self, c)
                                self.actions.append(widgetInstance)
                                log_manager.info(
                                    f'Widget uri registered from "{c.__module__}"'
                                )
                            else:
                                log_manager.info(
                                    f'Widget disabled and not registered from "{c.__module__}"'
                                )

        except Exception:
            pass

    def add_permission(self, name, permission):
        from .database import Permission
        from aurori import db
        key = self.name + "." + name
        description = "No description available"

        if hasattr(permission, 'description'):
            description = permission.description

        engine = db.engine
        table_exists = sa.inspect(engine).has_table(Permission.__tablename__)
        if table_exists:
            with db.get_session() as db_session:
                p = db_session.query(Permission).filter_by(name=key).first()
                if p is None:
                    p = Permission(name=key, description=description)
                    db_session.add(p)
                    db_session.commit()
                    p = db_session.query(Permission).filter_by(name=key).first()
                else:
                    p.caption = description
                    db_session.commit()
                self.permissions[key] = p
        else:
            log_manager.error(
                'Unable to create permission "{}" for feature {}'.format(
                    name, self.name))

    def getPermission(self, name):
        key = self.name + "." + name
        return self.permissions[key]



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import sys
import collections
import datetime
import threading
import logging
from .logger import Logger

class LogManager(object):
    def __init__(self):
        gettrace = getattr(sys, 'gettrace', None)
        if gettrace and gettrace():
            self.run_in_debug_mode = True
        else:
            self.run_in_debug_mode = False

        self.log_queue = collections.deque(maxlen=512)
        self.loggers = {
            "core": Logger("core", self.run_in_debug_mode, app_logger=False)
        }
        self.logger = self.loggers["core"]

        logging.basicConfig(level=logging.INFO, format='[%(asctime)s]-[%(name)s]-[%(levelname)s]: %(message)s', datefmt="%Y-%m-%dT%H:%M:%S")

    def register_logger(self, logger, name):
        self.loggers[name.lower()] = logger

    def info(self, msg, *args, **kwargs):
        """
        Delegate a info log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg.format(*args)
            indents = s.replace("\n", "\n" + " " * 29)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [INFO] : " + indents)
        else:
            s = str(msg)
            indents = s.replace("\n", "\n" + " " * 29)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [INFO] : " + indents)
        try:
            self.logger.info(msg, *args, **kwargs)
            indents = s.replace("\n", "\n" + " " * 14)
        except Exception as e:
            print("logger failed to print: ", s, e)

    def warning(self, msg, *args, **kwargs):
        """
        Delegate a warning log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [WARN] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [WARN] : " + s)

        try:
            self.logger.warning(msg, *args, **kwargs)
        except Exception:
            print("logger failed to print: ", s)

    def error(self, msg, *args, **kwargs):
        """
        Delegate a error log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [FAIL] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [FAIL] : " + s)

        try:
            self.logger.error(msg, *args, **kwargs)
        except Exception:
            print("logger failed to print: ", s)

    def debug(self, msg, *args, **kwargs):
        """
        Delegate a error log call to the underlying logger,
        """
        if self.run_in_debug_mode is not True:
            return
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [DEBUG] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [DEBUG] : " + s)

        try:
            self.logger.debug(msg, *args, **kwargs)
        except Exception:
            print("logger failed to print: ", s)



"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

from aurori.logs import log_manager
import hashlib


class UserManager(object):
    """The UserManager ..."""

    def __init__(
        self,
    ):
        # preparation to instanciate
        pass

    def init_manager(self, app, db, feature_manager, config):
        self.config = config
        self.app = app
        self.db = db
        self.feature_manager = feature_manager
        log_manager.info("UserManager initialized")

        try:
            __import__("ldap")
            log_manager.info("Import ldap")
        except ImportError:
            log_manager.info("Unable to import ldap")

        self.user = feature_manager.user_class

    def remove_user(self, email):
        u = self.user.query.filter_by(email=email).first()
        if u is not None:
            self.db.session.delete(u)
            self.db.session.commit()

    def register_user(self, userdata, ldap=False):
        if self.checkUserExist(userdata["email"]):
            return None
        else:
            with self.db.get_session() as db_session:
                u = self.user(
                    email=userdata["email"],
                    password=userdata["password"],
                    isAdmin=False,
                )
                u.firstname = userdata["firstname"]
                u.lastname = userdata["lastname"]
                u.organization = userdata["organization"]
                u.account_activated = True
                u.ldap = ldap
                if ldap is True:
                    u.ldap_dn = userdata["ldap_dn"]
                    u.ldap_username = userdata["ldap_username"]
                db_session.add(u)
                db_session.commit()
                return u

    def registerLdapUser(self, email):
        print("")

    def checkLdapUser(self, userid):
        import ldap

        server = self.config["LDAP"].get("server")
        bind_dn = self.config["LDAP"].get("bind_dn")
        bind_pw = self.config["LDAP"].get("bind_pw")
        search_dn = self.config["LDAP"].get("user_dn")
        default_organization = self.config["LDAP"].get("default_organization", "")

        attrs = [
            "givenName",
            "sn",
            "cn",
            "userPrincipalName",
            "mail",
            "sAMAccountName",
            "displayName",
        ]

        ldap_con = ldap.initialize(server)
        ldap_con.simple_bind_s(bind_dn, bind_pw)
        fil = "(| (mail=*" + str(userid) + "*) (sAMAccountName=*" + str(userid) + "*))"
        dn, user = ldap_con.search_s(search_dn, ldap.SCOPE_SUBTREE, fil, attrs)[0]

        userdata = {
            "email": user["mail"][0].decode("utf-8"),
            "password": "ldap",
            "firstname": user["givenName"][0].decode("utf-8"),
            "lastname": user["sn"][0].decode("utf-8"),
            "organization": default_organization,
            "ldap_dn": dn,
            "ldap_username": user["sAMAccountName"][0].decode("utf-8"),
        }
        u = self.register_user(userdata, True)
        return u

    def get_user_by_api_key(self, plaintext_api_key):
        if plaintext_api_key is not None:
            api_key_hash = hashlib.sha512(plaintext_api_key.encode("utf8")).hexdigest()
            return self.user.query.filter_by(api_key=api_key_hash).first()
        else:
            return None

    def getUser(self, userid):
        with self.db.get_session() as db_session:
            user = db_session.query(self.user).filter_by(email=userid).first()
            if userid is not None:
                if user is None and self.config["LDAP"].get("enable", False):
                    print("Search for ldap user locally")
                    user = (
                        db_session.query(self.user)
                        .filter_by(ldap_username=userid)
                        .first()
                    )
                    if user is None:
                        print("Search for ldap user on server")
                        try:
                            self.checkLdapUser(userid)
                            user = (
                                db_session.query(self.user)
                                .filter_by(ldap_username=userid)
                                .first()
                            )
                        except Exception as e:
                            print(e)
                            user = None
            if user and user.account_locked:
                return None
            else:
                return user

    def checkUserExist(self, email):
        with self.db.get_session() as db_session:
            user = db_session.query(self.user).filter_by(email=email).first()
            if user is None:
                return False
            else:
                return True

    def checkUserPassword(self, user, password):
        if user.ldap is False:
            return user.checkPassword(password)
        else:
            import ldap

            server = self.config["LDAP"].get("server")
            try:
                ldap_con = ldap.initialize(server)
                ldap_con.simple_bind_s(user.ldap_dn, password)
                return True
            except Exception as e:
                print("ldap error", e)
                return False



import sys
import threading
import collections
import datetime
import logging


class Logger(object):
    def __init__(self, name, run_in_debug_mode= False, app_logger=True):
        self.run_in_debug_mode = run_in_debug_mode
        if app_logger:
            name = "APP." + ".".join(name.split(".")[1:])
        self.logger = logging.getLogger(f'{name.upper()}')
        self.log_queue = collections.deque(maxlen=512)

    def info(self, msg, *args, **kwargs):
        """
        Delegate a info log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg.format(*args)
            indents = s.replace("\n", "\n" + " " * 29)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [INFO] : " + indents)
        else:
            s = str(msg)
            indents = s.replace("\n", "\n" + " " * 29)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [INFO] : " + indents)
        try:
            self.logger.info(msg)
            indents = s.replace("\n", "\n" + " " * 14)
        except Exception as e:
            print("logger failed to print: ", s, e)

    def warning(self, msg, *args):
        """
        Delegate a warning log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [WARN] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [WARN] : " + s)

        try:
            self.logger.warn(msg)
        except Exception:
            print("logger failed to print: ", s)

    def error(self, msg, *args, **kwargs):
        """
        Delegate a error log call to the underlying logger,
        """
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [FAIL] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [FAIL] : " + s)

        try:
            self.logger.error(msg)
        except Exception:
            print("logger failed to print: ", s)

    def debug(self, msg, *args, **kwargs):
        """
        Delegate a error log call to the underlying logger,
        """
        if self.run_in_debug_mode is not True:
            return
        if type(msg) is str:
            try:
                s = msg % args
            except Exception:
                s = msg
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [DEBUG] : " + s)
        else:
            s = str(msg)
            self.log_queue.append(
                datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S") +
                " [DEBUG] : " + s)

        try:
            self.logger.debug(msg)
        except Exception:
            print("logger failed to print: ", s)




"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"
"""Base class to create and handle pages in the features
"""


class Page(object):
    title = 'Missing title'  # Shown label of the page in the menu
    group = None  # groupname multiple pages
    icon = 'warning'  # icon (in typeset of material design icons)
    route = '/error'  # routing
    builder = 'frontend'  # page get build by the client (frontend)
    rank = 0.0  # ranks (double) the page higher values are at the top of the menu
    # groups will be ranked by the sum of the rank-values of their entries
    requireLogin = True  # login is required to view the page
    requireAdmin = True  # admin is required to view the page
    requirePermission = None  # a permission is required in the meaning of one of the following

    disable = False

    def __init__(self):
        if not hasattr(self, 'name'):
            self.name = self.__class__.__name__



import click

from aurori.version import version as aurori_version

@click.group()
def core():
    pass

@core.command()
def version():
    """Display the current version."""
    click.echo(aurori_version)


"""
The aurori project

Copyright (C) 2022  Marcus Drobisch,

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

__authors__ = ["Marcus Drobisch"]
__contact__ = "aurori@fabba.space"
__credits__ = []
__license__ = "AGPLv3+"

import logging
import sys
from fastapi import APIRouter, FastAPI
from fastapi.middleware.cors import CORSMiddleware

from os import environ

from sqlalchemy_utils import database_exists

from aurori.config import configure_app, config_manager
from aurori.version import version
from aurori.logs import log_manager
from aurori.features import feature_manager
from aurori.users import userManager, userbase
from aurori.features.permission import Permission
from aurori.jobs import job_manager
from aurori.database import db, SQLModelBase
from aurori.utils import unicode_string_to_bytes
from aurori.pages import menu_builder

__version__ = version


def create_app(config, fastapi_config=None, minimal=False, api_prefix="", title="Aurori App", version="0.1.0"):
    config_manager.init_manager(config, fastapi_config)
    app = FastAPI(title=title, version=version)

    origins = [
        "*"
    ]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    db.init_app(config)



    from aurori import app_router
    app.include_router(app_router, tags=["Core"])

    from aurori.jobs import jobs_router
    app.include_router(jobs_router,prefix=api_prefix, tags=["Jobs"])

    from aurori.users import user_router, core_router  # noqa: F401, F811
    app.include_router(core_router, tags=["Core"])
    app.include_router(user_router, prefix=api_prefix + "/user", tags=["App.User"])

    if not minimal:
        job_manager.init_manager(config_manager.config)
        menu_builder.init_builder(app, db, userManager, feature_manager)

    feature_manager.init_app(app, config_manager.config, minimal=minimal,api_prefix=api_prefix)
    if not minimal:
        userManager.init_manager(app, db, feature_manager, config_manager.config)


    database_is_new = not database_exists(db.url)
    if database_is_new == True:
        log_manager.warning("No database found at", db.url)
        log_manager.warning("Creating new empty database")
        SQLModelBase.metadata.create_all(bind=db.engine)

    return app


# declare app routes
app_router = APIRouter()
from aurori import routes



from typing import Annotated
from fastapi import Depends, HTTPException, status
from fastapi.security import OAuth2PasswordBearer
import jwt
from aurori.users import userManager
from aurori.config import config_manager
from .typemodels import Token, TokenData

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="/token", auto_error=False)


async def get_manadatory_user(token: Annotated[str, Depends(oauth2_scheme)]):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    secret = config_manager.config["SECURITY"]["secret_key"]
    algorithm = config_manager.config["SECURITY"]["jwt_algorithm"]

    try:
        payload = jwt.decode(token, secret, algorithms=[algorithm])
        username: str = payload.get("sub")
        if username is None:
            raise credentials_exception
        token_data = TokenData(username=username)
    except jwt.PyJWTError:
        raise credentials_exception
    user = userManager.getUser(
        token_data.username
    )  # get_user(fake_users_db, username=token_data.username)
    if user is None:
        raise credentials_exception
    return user


async def get_optional_user(token: Annotated[str, Depends(oauth2_scheme)]):
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate credentials",
        headers={"WWW-Authenticate": "Bearer"},
    )
    secret = config_manager.config["SECURITY"]["secret_key"]
    algorithm = config_manager.config["SECURITY"]["jwt_algorithm"]

    try:
        payload = jwt.decode(token, secret, algorithms=[algorithm])
        username: str = payload.get("sub")
        if username is None:
            token_data = None
        else:
            token_data = TokenData(username=username)
    except jwt.PyJWTError:
        token_data = None
    if token_data is not None:
        user = userManager.getUser(
            token_data.username
        )  # get_user(fake_users_db, username=token_data.username)
        return user
    return None
